#  LLM Tracer
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass

import pydantic

from llmtracer.object_converter import DynamicObjectConverter, ObjectConverter


def test_object_converter():
    # create an ObjectConverter
    converter = DynamicObjectConverter()

    # add a converter for a dataclass
    @dataclass
    class Test:
        a: int
        b: str

    # Check that a vanilla dataclass is converted automatically
    assert converter(Test(1, '2')) == {'a': 1, 'b': '2'}

    def convert_test(test: Test, converter: ObjectConverter):
        return {'a': test.a + 1, 'b': test.b}

    converter.register_converter(convert_test, Test)

    # Check that the converter is used
    assert converter(Test(1, '2')) == {'a': 1 + 1, 'b': '2'}

    # add a converter for a Pydantic model
    @converter.add_converter()
    class TestModel(pydantic.BaseModel):
        a: int
        b: str

    # convert a Pydantic model
    assert converter(TestModel(a=1, b='2')) == {'a': 1, 'b': '2'}

    # convert a dict
    assert converter({'a': 1, 'b': '2'}) == {'a': 1, 'b': '2'}

    # convert an object that is not a dict, dataclass, or Pydantic model
    assert converter(TestModel) == repr(TestModel)

    # convert a list
    assert converter([1, 2, 3]) == [1, 2, 3]

    # convert a set
    assert converter({1, 2, 3}) == {1, 2, 3}

    # convert a tuple
    assert converter((1, 2, 3)) == (1, 2, 3)

    # convert a string
    assert converter('test') == 'test'

    # convert an int
    assert converter(1) == 1

    # convert a nested object
    assert converter({'a': Test(1, '2')}) == {'a': {'a': 2, 'b': '2'}}



#  LLM Tracer
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from contextlib import contextmanager

from llmtracer import module_filtering, trace_builder, trace_schema


def build_trace(
    module_filters: module_filtering.ModuleFiltersSpecifier | None = None,
    stack_frame_context: int = 3,
    name: str | None = None,
):
    """
    Context manager that allows to trace our program execution.
    """
    if not module_filters:
        module_filters = trace_builder.trace_module_filters

    builder = trace_builder.TraceBuilder(
        module_filters=module_filtering.module_filters(module_filters), stack_frame_context=stack_frame_context
    )
    builder.event_root.name = name
    return builder


def add_event(
    name: str,
    properties: dict[str, object] | None = None,
    kind: trace_schema.TraceNodeKind = trace_schema.TraceNodeKind.EVENT,
):
    """
    Add an event to the current scope.
    """
    current = trace_builder.TraceBuilder.get_current()
    if current is not None:
        current.add_event(name, properties, kind)


def register_object(obj: object, name: str, properties: dict[str, object]):
    """
    Register an object as unique, so that it will be serialized only once.
    """
    current = trace_builder.TraceBuilder.get_current()
    if current is not None:
        current.register_object(obj, name, properties)


def update_event_properties(properties: dict[str, object] | None = None, /, **kwargs):
    """
    Update the properties of the current event.
    """
    current = trace_builder.TraceBuilder.get_current()
    if current is not None:
        current.update_event_properties(properties, **kwargs)


def update_name(name: str):
    """
    Update the name of the current event.
    """
    current = trace_builder.TraceBuilder.get_current()
    if current is not None:
        current.update_name(name)


@contextmanager
def event_scope(
    name: str,
    properties: dict[str, object] | None = None,
    kind: trace_schema.TraceNodeKind = trace_schema.TraceNodeKind.SCOPE,
):
    """
    Context manager that allows to trace our program execution.
    """
    current = trace_builder.TraceBuilder.get_current()
    if current is None:
        yield
    else:
        with current.event_scope(name, properties=properties, kind=kind, skip_frames=2):
            yield



#  LLM Tracer
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json
import pprint  # noqa: F401
import threading
import typing
import weakref
from enum import Enum

import reflex as rx
import reflex_chakra as rc
from starlette import status

from llmtracer import Trace, TraceNode, TraceNodeKind

from .flame_graph import FlameGraphNode, flame_graph
from .json_view import json_view
from .pcconfig import config

docs_url = "https://pynecone.io/docs/getting-started/introduction"
filename = f"{config.app_name}/{config.app_name}.py"


# solarized colors as HTML hex
# https://ethanschoonover.com/solarized/
class SolarizedColors(str, Enum):
    base03 = "#002b36"
    base02 = "#073642"
    base01 = "#586e75"
    base00 = "#657b83"
    base0 = "#839496"
    base1 = "#93a1a1"
    base2 = "#eee8d5"
    base3 = "#fdf6e3"
    yellow = "#b58900"
    orange = "#cb4b16"
    red = "#dc322f"
    magenta = "#d33682"
    violet = "#6c71c4"
    blue = "#268bd2"
    cyan = "#2aa198"
    green = "#859900"


class NodeInfo(rx.Base):
    node_name: str
    kind: TraceNodeKind
    event_id: int
    start_time_ms: int
    end_time_ms: int
    # delta_frame_infos: list[dict[str, object]]
    properties: object | None
    exception: object | None
    self_object: object | None
    result: object | None
    arguments: object | None


class Wrapper(rx.Base):
    inner: object = None

    def __init__(self, inner):
        super().__init__(inner=inner)


def load_example_trace():
    return Trace.parse_file("optimization_unit_trace_example.json")


async def send_event(state: rx.State, event_handler: rx.event.EventHandler):
    """Send an event to the server (from a different client thread)."""
    events = rx.event.fix_events([event_handler], state.get_token())
    state_update = rx.state.StateUpdate(delta={}, events=events)
    state_update_json = state_update.json()

    await app.sio.emit(
        str(rx.constants.SocketEvent.EVENT),
        state_update_json,
        to=state.get_sid(),
        namespace="/event",
    )


class StreamedTracesSingleton:
    """A class that streams a trace to the client."""

    traces: dict[str, Trace] = {}
    _receivers: weakref.WeakValueDictionary[str, rx.State] = weakref.WeakValueDictionary()

    lock: threading.Lock = threading.Lock()

    def send_update_event(self, origin: rx.State | None, trace_name: str):
        """Send an update event to all registered states (except the one that triggered the update)."""
        if origin is not None:
            origin = origin.parent_state if origin.parent_state else origin

        async def send_events():
            # send an event
            for state in self._receivers.values():
                if origin is None or state.get_sid() != origin.get_sid():
                    # noinspection PyNoneFunctionAssignment,PyArgumentList
                    event_handler: rx.event.EventHandler = State.on_injected_trace(trace_name)  # type: ignore
                    print(f"Sending update event for trace {trace_name} to {state.get_sid()}")
                    await send_event(state, event_handler)

        print(f"Sending update event for trace {trace_name}")
        app.sio.start_background_task(send_events)

    def update_trace(self, trace_name: str, trace: Trace):
        """Update the trace with the given name."""
        with self.lock:
            print(f"Updating trace {trace_name}")
            self.traces[trace_name] = trace
            self.send_update_event(None, trace_name)

    def register_state(self, state: rx.State):
        """Register a state to receive update events.

        We use this to keep track of active clients using a weakref value dictionary.
        """
        main_state: rx.State = state.parent_state if state.parent_state else state
        # TODO: check that we have exactly one state per session id?
        if main_state.get_sid() not in self._receivers:
            self._receivers[main_state.get_sid()] = main_state


def convert_trace_node_kind_to_color(kind: TraceNodeKind):
    if kind == TraceNodeKind.SCOPE:
        return SolarizedColors.base1
    elif kind == TraceNodeKind.AGENT:
        return SolarizedColors.green
    elif kind == TraceNodeKind.LLM:
        return SolarizedColors.blue
    elif kind == TraceNodeKind.CHAIN:
        return SolarizedColors.cyan
    elif kind == TraceNodeKind.CALL:
        return SolarizedColors.yellow
    elif kind == TraceNodeKind.EVENT:
        return SolarizedColors.orange
    elif kind == TraceNodeKind.TOOL:
        return SolarizedColors.magenta
    else:
        return SolarizedColors.base2


def convert_node_to_color(node: TraceNode):
    if "exception" in node.properties:
        return SolarizedColors.red
    else:
        return "black"


def convert_trace_to_flame_graph_data(trace: Trace) -> dict:
    def convert_node(node: TraceNode, discount=1.0) -> FlameGraphNode:
        children = []
        last_ms = node.start_time_ms
        for child in node.children:
            gap_s = child.start_time_ms - last_ms
            if gap_s > 0:
                children.append(
                    FlameGraphNode(
                        name="",
                        background_color="#00000000",
                        value=gap_s,
                        children=[],
                    )
                )
            children.append(convert_node(child, discount=discount * 0.95))
            last_ms = child.end_time_ms

        duration_s = node.end_time_ms - node.start_time_ms
        node_name = node.name or "/Unnamed/"
        return FlameGraphNode(
            id=str(node.event_id),
            name=node_name if not node.running else f"{node_name} (*)",
            value=duration_s * discount,
            children=children,
            background_color=convert_trace_node_kind_to_color(node.kind),
            color=convert_node_to_color(node),
        )

    converted_node = convert_node(trace.traces[-1]).dict(exclude_unset=True)
    return converted_node


class State(rx.State):
    """The app state."""

    # __slots__ = [
    #     '__weakref__',
    #     'flame_graph_data',
    #     'current_node',
    #     'trace_name',
    #     'injected_trace_names',
    # ]

    flame_graph_data: dict = FlameGraphNode(name="", value=1, background_color="#00000000", children=[]).model_dump()
    current_node: list[NodeInfo] = []
    trace_name: str | None = ""

    _trace: Trace | None = None
    _event_id_map: dict[int, TraceNode] = {}

    injected_trace_names: list[str] = []

    # def __init__(self, *args, **kwargs):
    #     super().__init__(*args, **kwargs)
    #     self._trace = None
    #     self._event_id_map = {}

    def register_state(self):
        """Register this state to receive updates."""
        print("Registering state")
        streamed_traced_singleton.register_state(self)
        self.injected_trace_names = list(streamed_traced_singleton.traces.keys())

    def reset_graph(self):
        self._trace = None
        self.trace_name = None
        self.update_flame_graph()

    def load_default_flame_graph(self):
        print("Loading default flame graph")
        self._trace = load_example_trace()
        self.trace_name = None
        self.update_flame_graph()

    async def handle_trace_upload(self, files: list[rx.UploadFile]):
        """Handle the upload of a file.

        Args:
            files: The uploaded file.
        """
        print(f"Received file upload {files}")
        assert len(files) == 1, "Expected exactly one file"
        file = files[0]
        upload_data = await file.read()
        self._trace = Trace.parse_raw(upload_data)  # type: ignore
        self.trace_name = None
        self.update_flame_graph()

    def update_flame_graph(self):
        if self._trace is None:
            self.flame_graph_data = FlameGraphNode(name="", value=1, background_color="#00000000", children=[]).dict()
            self._event_id_map = {}
        else:
            self._event_id_map = self._trace.build_event_id_map()
            self.flame_graph_data = convert_trace_to_flame_graph_data(self._trace)
        # is there a current node?
        if self.current_node:
            event_id = self.current_node[-1].event_id
            if event_id in self._event_id_map:
                self._set_current_node(event_id)
            else:
                self.current_node = []

    def on_injected_trace(self, trace_name: str):
        """Handle an injected trace event."""
        # parse trace_name as json
        trace_name = json.loads(trace_name)

        self.injected_trace_names = list(streamed_traced_singleton.traces.keys())

        if self._trace is None:
            self.trace_name = trace_name

        if self.trace_name == trace_name:
            self._trace = streamed_traced_singleton.traces.get(trace_name, None)  # type: ignore

            self.update_flame_graph()

    def handle_trace_selection(self, trace_name: str):
        """Handle the selection of a trace."""
        self.trace_name = trace_name
        self._trace = streamed_traced_singleton.traces.get(trace_name, None)  # type: ignore
        if self._trace is None:
            print(f"Could not find trace {trace_name}")

        self.update_flame_graph()

    def update_current_node(self, chart_data: dict):
        node_id = chart_data["source"].get("id", None)
        if node_id is None:
            event_id = None
        else:
            event_id = int(node_id)
        self._set_current_node(event_id)

    def _set_current_node(self, event_id: int | None):
        if event_id is None:
            self.current_node = []
        else:
            trace_node = self._event_id_map[event_id]

            properties = trace_node.properties.copy()
            exception = properties.pop("exception", None)
            arguments = properties.pop("arguments", None)
            result = properties.pop("result", None)
            if arguments:
                assert isinstance(arguments, dict)
                arguments = arguments.copy()
                self_object = arguments.pop("self", None)
            else:
                self_object = None

            optional_properties = properties if properties else None
            self.current_node = [
                NodeInfo(
                    node_name=trace_node.name,
                    kind=trace_node.kind,
                    event_id=trace_node.event_id,
                    start_time_ms=trace_node.start_time_ms,
                    end_time_ms=trace_node.end_time_ms,
                    properties=optional_properties,
                    exception=exception,
                    self_object=self_object,
                    result=result,
                    arguments=arguments,
                )
            ]


def render_node_info(node_info: NodeInfo):
    """Renders node_info as a simple table."""
    header_style = dict(width="20%", text_align="right", vertical_align="top")

    return rc.table_container(
        rc.table(
            rc.tbody(
                rc.tr(
                    rc.th("Name", style=header_style),
                    rc.td(node_info.node_name, colspan=0),
                ),
                rc.tr(
                    rc.th("", style=header_style),
                    rc.td(
                        rc.stat_group(
                            rc.stat(
                                rc.stat_number(node_info.kind),
                                rc.stat_help_text("KIND"),
                            ),
                            rc.stat(
                                rc.stat_number((node_info.end_time_ms - node_info.start_time_ms) / 1000),
                                rc.stat_help_text("DURATION (S)"),
                            ),
                            width="100%",
                        )
                    ),
                ),
                rx.cond(
                    node_info.exception,
                    rc.tr(
                        rc.th("Exception", style=header_style),
                        rc.td(
                            json_view(
                                data=node_info.exception,
                                style=dict(background_color="#000000"),
                            ),
                            colspan=0,
                        ),
                    ),
                ),
                rx.cond(
                    node_info.result,
                    rc.tr(
                        rc.th("Result", style=header_style),
                        rc.td(
                            json_view(
                                data=node_info.result,
                                style=dict(background_color="#000000"),
                            ),
                            colspan=0,
                        ),
                    ),
                ),
                rx.cond(
                    node_info.arguments,
                    rc.tr(
                        rc.th("Arguments", style=header_style),
                        rc.td(
                            json_view(
                                data=node_info.arguments,
                                style=dict(background_color="#000000"),
                            ),
                            colspan=0,
                        ),
                    ),
                ),
                rx.cond(
                    node_info.self_object,
                    rc.tr(
                        rc.th("Self", style=header_style),
                        rc.td(
                            json_view(
                                data=node_info.self_object,
                                style=dict(background_color="#000000"),
                            ),
                            colspan=0,
                        ),
                    ),
                ),
                rx.cond(
                    node_info.properties,
                    rc.tr(
                        rc.th("Properties", style=header_style),
                        rc.td(
                            json_view(
                                data=node_info.properties,
                                style=dict(background_color="#000000"),
                            ),
                            colspan=0,
                        ),
                    ),
                ),
            ),
            variant="simple",
        ),
        style=dict(width=1024),
    )


@typing.no_type_check
def index() -> rx.Component:
    return rc.center(
        rc.vstack(
            rc.span(
                rc.heading(
                    "Trace Viewer",
                    level=1,
                    style=dict(display="inline-block", margin_right="16px"),
                ),
                rx.popover.root(
                    rx.popover.trigger(rc.button(rc.icon(tag="hamburger"), style=dict(vertical_align="top"))),
                    rx.popover.content(
                        rx.flex(
                            rc.heading("Choose Trace Source"),
                            rc.button("Load Example", on_click=State.load_default_flame_graph),
                            rc.divider(margin="0.5em"),
                            rx.upload(
                                rc.text("Drag and drop files here or click to select trace json file."),
                                rx.vstack(rx.foreach(rx.selected_files("trace-upload"), rx.text)),
                                border="1px dotted rgb(107,99,246)",
                                padding="1em",
                                id="trace-upload",
                            ),
                            rc.button(
                                "Load",
                                on_click=lambda: State.handle_trace_upload(rx.upload_files(upload_id="trace-upload")),
                            ),
                            rc.divider(margin="0.5em"),
                            rc.select(
                                State.injected_trace_names,
                                is_disabled=State.injected_trace_names.length() == 0,
                                value=State.trace_name,
                                placeholder="Select an available trace",
                                on_change=State.handle_trace_selection,
                            ),
                            rc.divider(margin="0.5em"),
                            rc.button("Reset", on_click=State.reset_graph),
                            direction="column",
                            spacing="3",
                        )
                    ),
                ),
            ),
            rx.cond(
                State.flame_graph_data,
                rx.fragment(
                    rx.box(
                        flame_graph(
                            width=1024,
                            height=256,
                            data=State.flame_graph_data,
                            on_change=lambda data: State.update_current_node(data),  # type: ignore
                        ),
                        border="1px orange solid",
                    ),
                    rx.foreach(
                        State.current_node,
                        render_node_info,
                    ),
                ),
            ),
            # rx.cond(
            #     State.flame_graph_data,
            #     rx.fragment(
            #         rx.box(
            #             flame_graph(
            #                 width=1024,
            #                 height=256,
            #                 data=State.flame_graph_data,
            #                 on_change=lambda data: State.update_current_node(data),  # type: ignore
            #             ),
            #             border="1px orange solid",
            #         ),
            #         rx.foreach(
            #             State.current_node,
            #             render_node_info,
            #         ),
            #     ),
            # ),
        ),
        padding_top="64px",
        padding_bottom="64px",
    )


streamed_traced_singleton = StreamedTracesSingleton()

meta = [
    {"name": "theme_color", "content": SolarizedColors.green},
    {"char_set": "UTF-8"},
    # {"property": "og:url", "content": "url"},
]

# Add state and page to the app.
app = rx.App(
    stylesheets=[
        "react-json-view-lite.css",
    ],
    theme=rx.theme(
        appearance="light",
    ),
)
app.add_page(
    index,
    # on_load=State.register_state,
    meta=meta,
    title="Trace Viewer",
    description="View traces (both offline and streamed).",
)


@app.api.post("/trace/{trace_name}", status_code=status.HTTP_204_NO_CONTENT)
async def inject_trace(trace_name: str, trace: Trace):
    streamed_traced_singleton.update_trace(trace_name, trace)



#  LLM Tracer
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import reflex as rx

"""
https://monojack.github.io/react-object-view/

import { ObjectView } from 'react-object-view'

const options = {
  hideDataTypes: false,
  hideObjectSize: false,
  hidePreviews: false
}
const palette = {}
const styles = {}
const data = {/** copy this from the `Data` tab */}

function App() {
  return (
    <ObjectView
      data={data}
      options={options}
      styles={styles}
      palette={palette}
    />
  )
}
"""


class JsonView(rx.Component):
    if False:
        library = "react-object-view"
        tag = "ObjectView"
    elif True:
        library = "react-json-view-lite"
        tag = "JsonView"
    data: rx.Var[object]
    # options: pc.Var[dict] = {}
    # styles: pc.Var[dict] = {}
    # palette: pc.Var[dict] = {}

    # @classmethod
    # def get_controlled_triggers(cls) -> dict[str, pc.Var]:
    #     return {"on_change": pc.EVENT_ARG}


json_view = JsonView.create



"""
Simple logger/execution tracker that uses tracks the stack frames and 'data'.
"""
#  LLM Tracer
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import inspect
import time
import traceback
import typing
from contextlib import contextmanager
from contextvars import ContextVar
from dataclasses import dataclass, field
from functools import partial, wraps
from typing import ClassVar

from langchain.schema import BaseMessage

from llmtracer import module_filtering
from llmtracer.frame_info import FrameInfo, get_frame_infos
from llmtracer.object_converter import DynamicObjectConverter, ObjectConverter, convert_pydantic_model
from llmtracer.trace_schema import Trace, TraceNode, TraceNodeKind
from llmtracer.utils.callable_wrapper import CallableWrapper
from llmtracer.utils.weakrefs import WeakKeyIdMap

T = typing.TypeVar("T")
P = typing.ParamSpec("P")


trace_object_converter = DynamicObjectConverter()
trace_module_filters = None

# TODO: move this somewhere else?
# chat messages need to be converted to JSON
trace_object_converter.register_converter(convert_pydantic_model, BaseMessage)


def default_timer() -> int:
    """
    Default timer for the tracer.

    Returns:
        The current time in milliseconds.
    """
    return int(time.time() * 1000)


@dataclass
class TraceNodeBuilder:
    """
    A node builder in the trace tree.
    """

    kind: TraceNodeKind
    name: str | None
    event_id: int
    start_time_ms: int
    delta_frame_infos: list[FrameInfo]
    stack_height: int

    end_time_ms: int | None = None
    parent: 'TraceNodeBuilder | None' = None
    children: list['TraceNodeBuilder'] = field(default_factory=list)
    properties: dict[str, object] = field(default_factory=dict)

    @classmethod
    def create_root(cls):
        return cls(
            kind=TraceNodeKind.SCOPE,
            name=None,
            event_id=0,
            start_time_ms=0,
            delta_frame_infos=[],
            stack_height=0,
        )

    def get_delta_frame_infos(
        self, num_frames_to_skip: int = 0, module_filters: module_filtering.ModuleFilters | None = None, context=3
    ):
        frame_infos, full_stack_height = get_frame_infos(
            num_top_frames_to_skip=num_frames_to_skip + 1,
            num_bottom_frames_to_skip=self.stack_height,
            module_filters=module_filters,
            context=context,
        )

        return frame_infos, full_stack_height

    def build(self):
        return TraceNode(
            kind=self.kind,
            name=self.name,
            event_id=self.event_id,
            start_time_ms=self.start_time_ms,
            end_time_ms=self.end_time_ms or default_timer(),
            running=self.end_time_ms is None,
            delta_frame_infos=self.delta_frame_infos,
            properties=self.properties,
            children=[sub_event.build() for sub_event in self.children],
        )


class TraceBuilderEventHandler:
    def on_scope_final(self, builder: 'TraceBuilder'):
        pass

    def on_event_scope_final(self, builder: 'TraceBuilder'):
        pass


@dataclass(weakref_slot=True, slots=True)
class TraceBuilder:
    _current: ClassVar[ContextVar['TraceBuilder | None']] = ContextVar("current_trace_builder", default=None)

    module_filters: module_filtering.ModuleFilters
    stack_frame_context: int

    event_root: TraceNodeBuilder = field(default_factory=TraceNodeBuilder.create_root)
    object_map: WeakKeyIdMap[object, str] = field(default_factory=WeakKeyIdMap)
    unique_objects: dict[str, dict] = field(default_factory=dict)

    id_counter: int = 0
    current_event_node: TraceNodeBuilder | None = None

    event_handlers: list[TraceBuilderEventHandler] = field(default_factory=list)

    def build(self):
        return Trace(
            name=self.event_root.name,
            properties=self.event_root.properties,
            traces=[child.build() for child in self.event_root.children],
            unique_objects=self.unique_objects,
        )

    def next_id(self):
        self.id_counter += 1
        return self.id_counter

    @contextmanager
    def scope(self, name: str | None = None):
        """
        Context manager that allows to trace our program execution.
        """
        assert self.current_event_node is None
        self.current_event_node = self.event_root

        token = self._current.set(self)
        try:
            with self.event_scope(name=name, kind=TraceNodeKind.SCOPE, skip_frames=2):
                yield self
        finally:
            for handler in self.event_handlers:
                handler.on_scope_final(self)
            self._current.reset(token)
            self.current_event_node = None

    @contextmanager
    def event_scope(
        self,
        name: str | None,
        properties: dict[str, object] | None = None,
        kind: TraceNodeKind = TraceNodeKind.SCOPE,
        skip_frames: int = 0,
    ):
        """
        Context manager that allows to trace our program execution.
        """
        assert self._current.get() is self
        assert self.current_event_node is not None

        if properties is None:
            properties = {}

        start_time = default_timer()
        delta_frame_infos, stack_height = self.current_event_node.get_delta_frame_infos(
            num_frames_to_skip=2 + skip_frames, module_filters=self.module_filters, context=self.stack_frame_context
        )
        event_node = TraceNodeBuilder(
            kind=kind,
            name=name,
            event_id=self.next_id(),
            start_time_ms=start_time,
            delta_frame_infos=delta_frame_infos,
            stack_height=stack_height - 1,
            parent=self.current_event_node,
            properties=dict(properties),
        )
        self.current_event_node.children.append(event_node)

        old_event_node = self.current_event_node
        self.current_event_node = event_node

        try:
            yield
        except BaseException as e:
            self.update_event_properties(exception='\n'.join(traceback.TracebackException.from_exception(e).format()))
            raise
        finally:
            event_node.end_time_ms = default_timer()
            self.current_event_node = old_event_node

            for handler in self.event_handlers:
                handler.on_event_scope_final(self)

    def register_object(self, obj: object, name: str, properties: dict[str, object]):
        # Make name unique if needed
        if name in self.unique_objects:
            # if we are in a scope, we can use the scope name as a prefix
            if self.current_event_node is not None:
                name = f"{self.current_event_node.name}_{name}"

            if name in self.unique_objects:
                i = 1
                while f"{name}[{i}]" in self.unique_objects:
                    i += 1
                name = f"{name}[{i}]"
        self.object_map[obj] = name
        self.unique_objects[name] = properties

    def convert_object(self, obj: object, preferred_object_converter: ObjectConverter | None = None):
        if preferred_object_converter is None:
            preferred_object_converter = self.convert_object

        # if the object is in the map, we return its name as a reference
        if obj in self.object_map:
            return dict(unique_object=self.object_map[obj])

        return trace_object_converter(obj, preferred_object_converter)

    @classmethod
    def get_current(cls) -> 'TraceBuilder | None':
        return cls._current.get()

    @classmethod
    def get_current_node(cls) -> 'TraceNodeBuilder | None':
        current = cls.get_current()
        if current is None:
            return None
        else:
            return current.current_event_node

    def add_event(
        self,
        name: str,
        properties: dict[str, object] | None = None,
        kind: TraceNodeKind = TraceNodeKind.EVENT,
    ):
        """
        Add an event to the current scope.
        """
        if properties is None:
            properties = {}
        with self.event_scope(name, properties=properties, kind=kind, skip_frames=2):
            pass

    def update_event_properties(self, properties: dict[str, object] | None = None, /, **kwargs):
        """
        Update the properties of the current event.
        """
        assert self.current_event_node is not None
        if properties is None:
            properties = {}
        self.current_event_node.properties.update(self.convert_object(properties | kwargs))

    def update_name(self, name: str):
        """
        Update the name of the current event.
        """
        assert self.current_event_node is not None
        self.current_event_node.name = name


@dataclass
class CallTracer(CallableWrapper, typing.Callable[P, T], typing.Generic[P, T]):  # type: ignore
    __signature__: inspect.Signature
    __wrapped__: typing.Callable[P, T]
    __wrapped_name__: str
    __kind__: TraceNodeKind = TraceNodeKind.CALL
    __capture_return__: bool = False
    __capture_args__: bool | list[str] | slice = False
    __object_converter__: DynamicObjectConverter | None = None

    def __call__(self, *args, **kwargs):
        # check if we are in a trace
        builder = TraceBuilder.get_current()
        if builder is None:
            return self.__wrapped__(*args, **kwargs)

        object_converter = self.__object_converter__
        if object_converter is None:
            object_converter = builder.convert_object

        # build properties
        properties = {}
        if self.__capture_args__ is not False:
            # bind the arguments to the signature
            bound_args = self.__signature__.bind(*args, **kwargs)
            # add the arguments to the properties
            if self.__capture_args__ is True:
                arguments = bound_args.arguments
            elif isinstance(self.__capture_args__, list):
                arguments = {arg: bound_args.arguments[arg] for arg in self.__capture_args__}
            elif isinstance(self.__capture_args__, slice):
                arguments = {
                    arg: bound_args.arguments[arg] for arg in list(bound_args.arguments)[self.__capture_args__]
                }

            # anything that can be stored in a json is okay
            converted_arguments = {}
            for arg, value in arguments.items():
                converted_arguments[arg] = object_converter(value)
            properties["arguments"] = converted_arguments

        # create event scope
        with builder.event_scope(self.__wrapped_name__, properties, kind=self.__kind__, skip_frames=1):
            # call the function
            result = self.__wrapped__(*args, **kwargs)

            if self.__capture_return__:
                builder.current_event_node.properties.update({"result": object_converter(result)})
        return result


class Slicer:
    def __class_getitem__(cls, item):
        return item


slicer = Slicer


def trace_calls(
    func=None,
    *,
    name: str | None = None,
    kind: TraceNodeKind = TraceNodeKind.CALL,
    capture_return: bool = False,
    capture_args: bool | list[str] | slice = False,
    object_converter: DynamicObjectConverter | None = None,
):
    """
    Decorator that allows to trace our program execution.
    """
    if func is None:
        return partial(
            trace_calls,
            name=name,
            kind=kind,
            capture_return=capture_return,
            capture_args=capture_args,
        )

    # get the signature of the function
    signature = inspect.signature(func)

    # if capture_args is an iterable, convert it to a set
    if isinstance(capture_args, typing.Iterable):
        assert not isinstance(capture_args, bool)
        arg_names = set(capture_args)

        # check that all the arguments are valid
        for arg in arg_names:
            if arg not in signature.parameters:
                raise ValueError(f"Argument '{arg}' is not a valid argument of function '{func.__name__}'!")

    # get the name of the function
    if name is None:
        name = func.__name__

    wrapped_function = wraps(func)(
        CallTracer(
            __signature__=signature,
            __wrapped__=func,
            __wrapped_name__=name,
            __kind__=kind,
            __capture_return__=capture_return,
            __capture_args__=capture_args,
            __object_converter__=object_converter,
        )
    )

    return wrapped_function



#  LLM Tracer
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from llmtracer.tools.trace_viewer.app.endpoint_integration import trace_viewer_send_trace_builder
from llmtracer.trace_builder import TraceBuilder, TraceBuilderEventHandler


class TraceViewerIntegration(TraceBuilderEventHandler):
    def on_scope_final(self, builder: 'TraceBuilder'):
        trace_viewer_send_trace_builder(builder, force=True)

    def on_event_scope_final(self, builder: 'TraceBuilder'):
        trace_viewer_send_trace_builder(builder, force=False)



#  LLM Tracer
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import types
import typing
from dataclasses import dataclass

ModuleSpecifier: typing.TypeAlias = types.ModuleType | str
ModuleFilterSpecifier: typing.TypeAlias = typing.Union[ModuleSpecifier, typing.Callable[[str], bool]]
ModuleFilterIterableSpecifier: typing.TypeAlias = typing.Iterable | ModuleFilterSpecifier | None


@dataclass
class OnlyModuleFilter:
    """
    A filter for a module. The filter matches the full module name.
    """

    module_name: str

    def __init__(self, module: ModuleSpecifier):
        if isinstance(module, str):
            self.module_name = module
        elif isinstance(module, types.ModuleType):
            self.module_name = module.__name__
        else:
            raise ValueError(f"Unknown module type {type(module)}")

    def __call__(self, target_module_name: str):
        return target_module_name == self.module_name


@dataclass
class ModulePrefixFilter:
    """
    A filter for a module. The filter matches the module name prefix.
    """

    module_name_prefix: str

    def __init__(self, module_prefix: ModuleSpecifier):
        if isinstance(module_prefix, str):
            self.module_name_prefix = module_prefix
        elif isinstance(module_prefix, types.ModuleType):
            self.module_name_prefix = module_prefix.__name__
        else:
            raise ValueError(f"Unknown module type {type(module_prefix)}")

    def __call__(self, target_module_name: str):
        return target_module_name.startswith(self.module_name_prefix)


def module_filter(module: ModuleSpecifier):
    if isinstance(module, types.ModuleType):
        return OnlyModuleFilter(module.__name__)
    elif isinstance(module, str):
        if module.endswith("*"):
            return ModulePrefixFilter(module[:-1])
        return OnlyModuleFilter(module)
    else:
        raise ValueError(f"Unknown filter type {type(module)} for module {module}")


def convert_module_filter_specifier_to_filter(module_filter_specifier: ModuleFilterSpecifier):
    if isinstance(module_filter_specifier, ModuleSpecifier):  # type: ignore
        return module_filter(module_filter_specifier)  # type: ignore
    elif callable(module_filter_specifier):
        return module_filter_specifier
    else:
        raise ValueError(f"Unknown filter type {type(module_filter_specifier)} for module {module_filter_specifier}")


def convert_module_filter_specifiers_to_filters(module_filter_specifiers: ModuleFilterIterableSpecifier):
    if module_filter_specifiers is None:
        return None
    elif not isinstance(module_filter_specifiers, str) and isinstance(module_filter_specifiers, typing.Iterable):
        assert isinstance(module_filter_specifiers, typing.Iterable)
        return [
            convert_module_filter_specifier_to_filter(module_filter_specifier)
            for module_filter_specifier in module_filter_specifiers  # type: ignore
        ]
    elif isinstance(module_filter_specifiers, ModuleSpecifier):  # type: ignore
        return [module_filter(module_filter_specifiers)]  # type: ignore
    elif callable(module_filter_specifiers):
        return [module_filter_specifiers]
    else:
        raise ValueError(f"Unknown filter type {type(module_filter_specifiers)} for module {module_filter_specifiers}")


def get_top_level_package(module_name: str):
    """
    Returns the top level package name for the given module name.
    """
    return module_name.split(".")[0]


@dataclass
class ModuleFilters:
    """
    Supports filtering of modules using a whitelist (or all) and a blacklist (or none).
    """

    include: list[typing.Callable]
    exclude: list[typing.Callable]

    @classmethod
    def create(cls, include: ModuleFilterIterableSpecifier = None, exclude: ModuleFilterIterableSpecifier = None):
        return cls(
            convert_module_filter_specifiers_to_filters(include), convert_module_filter_specifiers_to_filters(exclude)
        )

    def __call__(self, target_module_name: str):
        if self.include is not None:
            if not any(f(target_module_name) for f in self.include):
                return False
        if self.exclude is not None:
            if any(f(target_module_name) for f in self.exclude):
                return False
        return True


ModuleFiltersSpecifier = ModuleFilterIterableSpecifier | ModuleFilters


def module_filters(module_filters_specifier: ModuleFiltersSpecifier):
    if isinstance(module_filters_specifier, ModuleFilters):
        return module_filters_specifier
    else:
        return ModuleFilters.create(module_filters_specifier)



#  LLM Tracer
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import reflex as rx
from pydantic import BaseModel, Field


class FlameGraphNode(BaseModel):
    name: str
    value: int | float
    children: list["FlameGraphNode"]
    color: str | None = None
    backgroundColor: str | None = Field(None, alias="background_color")
    tooltip: str | None = None
    id: str | None = None


class FlameGraph(rx.Component):
    library = "react-flame-graph"
    tag = "FlameGraph"

    data: rx.Var[dict]
    height: rx.Var[int]
    width: rx.Var[int]

    on_change: rx.EventHandler[lambda chart_data: [chart_data]]


flame_graph = FlameGraph.create



#  LLM Tracer
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from contextlib import contextmanager

import wandb.sdk.wandb_run
from wandb.sdk.data_types import trace_tree

from llmtracer import (  # type: ignore
    Trace,
    TraceBuilder,
    TraceBuilderEventHandler,
    TraceNode,
    TraceNodeKind,
    build_trace,
    module_filtering,
)


def convert_event_kind_str(kind: TraceNodeKind):
    try:
        span_kind = trace_tree.SpanKind(kind.value)
    except ValueError:
        span_kind = trace_tree.SpanKind.AGENT

    return span_kind


def convert_result(result: object):
    if result is None:
        return None
    if isinstance(result, dict | list | set):
        return result
    return [result]


def build_span(node: TraceNode):
    span = trace_tree.Span()
    span.span_id = str(node.event_id)
    span.name = node.name if node.name is not None else ''
    span.span_kind = convert_event_kind_str(node.kind)

    span.start_time_ms = node.start_time_ms
    span.end_time_ms = node.end_time_ms

    if "exception" not in node.properties:
        span.status_code = trace_tree.StatusCode.SUCCESS
    else:
        span.status_code = trace_tree.StatusCode.ERROR
        span.status_message = repr(node.properties["exception"])

    span.add_named_result(
        node.properties.get('arguments', {}), convert_result(node.properties.get('result', None))  # type: ignore
    )

    properties = dict(node.properties)
    if "arguments" in properties:
        del properties["arguments"]
    if "result" in properties:
        del properties["result"]

    span.attributes = dict(
        properties=properties,
        delta_stack=node.delta_frame_infos,
    )
    span.child_spans = [build_span(child) for child in node.children]
    return span


def wandb_build_trace_trees(trace: Trace):
    media_list = []
    for trace_instance in trace.traces:
        scope_span = build_span(trace_instance)

        media = trace_tree.WBTraceTree(root_span=scope_span, model_dict=trace.unique_objects)
        media_list.append(media)
    return media_list


class WandBIntegration(TraceBuilderEventHandler):
    def on_scope_final(self, builder: 'TraceBuilder'):
        for trace_artifact in wandb_build_trace_trees(builder.build()):
            wandb.log({"trace": trace_artifact})  # type: ignore


@contextmanager
def wandb_tracer(
    name: str | None,
    *,
    module_filters: module_filtering.ModuleFiltersSpecifier | None = None,
    stack_frame_context: int = 3,
    event_handlers: list[TraceBuilderEventHandler] = [],
):
    tb = build_trace(
        module_filters=module_filters, stack_frame_context=stack_frame_context, name=wandb.run.name  # type: ignore
    )  # type: ignore
    tb.event_handlers.append(WandBIntegration())
    tb.event_handlers.extend(event_handlers)

    try:
        with tb.scope(name) as builder:
            yield builder
    finally:
        wandb.finish()  # type: ignore



#  LLM Tracer
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import types
import typing
from dataclasses import dataclass

P = typing.ParamSpec("P")
T = typing.TypeVar("T")


@dataclass
class CallableWrapper:
    """
    A functor that wraps a callable and forwards all attributes to the wrapped callable.
    """

    __wrapped__: object

    def __get__(self, instance: object, owner: type | None = None) -> typing.Callable:
        """Support instance methods."""
        if instance is None:
            return self

        # Bind self to instance as MethodType
        return types.MethodType(self, instance)

    def __getattr__(self, item):
        if item == "__wrapped__":
            raise AttributeError(name=item, obj=self)
        return getattr(self.__wrapped__, item)

    def __call__(self, *args, **kwargs):
        raise NotImplementedError

    def __dir__(self) -> typing.Iterable[str]:
        # merge the attributes of the wrapped object with the attributes of the wrapper
        self_dir = list(super().__dir__())
        wrapped_dir = list(dir(self.__wrapped__))
        return list(set(self_dir + wrapped_dir))


# Fix Pydantic (if it is installed) by adding CallableWrapper to UNTOUCHED_TYPES
try:
    import pydantic

    pydantic.BaseModel.model_config["ignored_types"] = pydantic.BaseModel.model_config.get("ignored_types", ()) + (
        CallableWrapper,
    )
except ImportError:
    pass



#  LLM Tracer
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import reflex as rx

config = rx.Config(
    app_name="app",
    db_url="sqlite:///pynecone.db",
    port=3333,
    backend_port=8333,
    api_url="http://localhost:8333",
    env=rx.Env.DEV,
)



#  LLM Tracer
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import inspect

from pydantic import BaseModel

from llmtracer import module_filtering


class FrameInfo(BaseModel):
    """
    A frame info object that is serializable.
    """

    module: str
    lineno: int
    function: str
    code_context: list[str] | None
    index: int | None
    # positions: dis.Positions | None = None


def get_frame_infos(
    num_top_frames_to_skip: int = 0,
    num_bottom_frames_to_skip: int = 0,
    module_filters: module_filtering.ModuleFilters | None = None,
    context: int = 3,
) -> tuple[list[FrameInfo], int]:
    # Get the current stack frame infos
    frame_infos: list[inspect.FrameInfo] = inspect.stack(context=context)

    # Get the stack frame infos for the delta stack summary
    caller_frame_infos = frame_infos[num_top_frames_to_skip + 1 :]

    stack_height = len(caller_frame_infos)

    # remove bottom frames
    relevant_inspect_frame_infos: list[inspect.FrameInfo] = caller_frame_infos[
        : stack_height - num_bottom_frames_to_skip
    ]

    relevant_frame_infos: list[FrameInfo] = [
        FrameInfo(
            module=module.__name__ if (module := inspect.getmodule(f.frame)) else "<unknown>",
            lineno=f.lineno,
            function=f.function,
            code_context=f.code_context,
            index=f.index,
            # positions=f.positions,
        )
        for f in relevant_inspect_frame_infos
    ]

    # Filter the stack frame infos
    if module_filters is not None:
        relevant_frame_infos = [f for f in relevant_frame_infos if module_filters(f.module)]

    return relevant_frame_infos, stack_height


def test_get_frame_infos():
    def f():
        frame_infos = get_frame_infos(module_filters=module_filtering.only_module(__name__))
        assert len(frame_infos) == 3
        assert frame_infos[0].function == "f"
        assert frame_infos[1].function == "g"
        assert frame_infos[2].function == "test_get_frame_infos"

    def g():
        f()

    g()



"""See `TestSet` for an example."""
#  LLM Tracer
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import MutableSet, Type

from llmtracer.testing.collections import unordered_equal


class MutableSetTests:
    mutable_set: Type

    @classmethod
    def create_mutable_set(cls) -> MutableSet:
        return cls.mutable_set()

    @staticmethod
    def get_element(i):
        return i

    def test_add_coverage(self):
        instance = self.create_mutable_set()
        element1 = self.get_element(1)

        instance.add(element1)
        # And add a second time for good measure
        instance.add(element1)

    def test_discard_missing_element_passes(self):
        instance = self.create_mutable_set()
        element1 = self.get_element(1)

        instance.discard(element1)

    def test_discard_passes(self):
        instance = self.create_mutable_set()
        element1 = self.get_element(1)

        instance.add(element1)
        instance.discard(element1)

    def test_contains_len(self):
        instance = self.create_mutable_set()
        element1 = self.get_element(1)

        assert len(instance) == 0
        assert element1 not in instance
        instance.add(element1)
        assert element1 in instance
        assert len(instance) == 1

        element2 = self.get_element(2)
        assert element2 not in instance
        instance.add(element2)
        assert element2 in instance
        assert len(instance) == 2

        assert element1 in instance
        instance.discard(element1)
        assert element1 not in instance
        assert len(instance) == 1

        assert element2 in instance
        instance.discard(element2)
        assert element1 not in instance
        assert element2 not in instance
        assert len(instance) == 0

    def test_iter(self):
        instance = self.create_mutable_set()
        element1 = self.get_element(1)
        element2 = self.get_element(2)

        assert list(iter(instance)) == []
        instance.add(element1)
        assert unordered_equal(iter(instance), [element1])

        instance.add(element2)
        assert unordered_equal(iter(instance), [element1, element2])



#  LLM Tracer
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import dataclasses
import inspect
import typing
from dataclasses import dataclass, field
from functools import partial

import pydantic

T = typing.TypeVar("T")

ObjectConverter: typing.TypeAlias = typing.Callable[[typing.Any, typing.Optional['ObjectConverter']], typing.Any]


@typing.no_type_check
def simple_object_converter(obj: typing.Any, preferred_converter: ObjectConverter | None = None) -> typing.Any:
    if preferred_converter is None:
        preferred_converter = simple_object_converter

    if isinstance(obj, pydantic.BaseModel):
        return convert_pydantic_model(obj, preferred_converter)
    elif not isinstance(obj, type) and dataclasses.is_dataclass(obj):
        return {preferred_converter(f.name): preferred_converter(getattr(obj, f.name)) for f in dataclasses.fields(obj)}
    elif isinstance(obj, (str, int, float, bool, type(None))):
        return obj
    elif isinstance(obj, tuple):
        return tuple(preferred_converter(v) for v in obj)
    elif isinstance(obj, list):
        return [preferred_converter(v) for v in obj]
    elif isinstance(obj, set):
        return {preferred_converter(v) for v in obj}
    elif isinstance(obj, dict):
        return {preferred_converter(k): preferred_converter(v) for k, v in obj.items()}

    return repr(obj)


@dataclass
class DynamicObjectConverter:
    """A class that converts objects to JSON dicts or strings.

    By default, only literals, tuples, lists, sets, dicts and dataclasses are converted.
    Other objects are {}.

    Additional classes can be added with the `add_converter` decorator.
    """

    default_converter: ObjectConverter = simple_object_converter
    converters: dict[type, typing.Callable[[typing.Any, ObjectConverter], typing.Any] | None] = field(
        default_factory=dict
    )

    def __call__(self, obj: object, preferred_converter: ObjectConverter | None = None):
        if preferred_converter is None:
            preferred_converter = self

        for t in reversed(self.converters):
            if not isinstance(obj, t):
                continue

            converter = self.converters[t]
            if converter is not None:
                return converter(obj, preferred_converter)
            else:
                return f'{obj.__class__.__module__}:{obj.__class__.__qualname__} @ {hex(id(obj))}'

        return self.default_converter(obj, preferred_converter)  # type: ignore

    def register_converter(
        self, func: typing.Callable[[T, ObjectConverter], dict] | None = None, type_: type[T] | None = None
    ):
        """
        Registers a converter for a type.

        This can be called as a decorator or as a function.
        """
        if func is None:
            # return a decorator that adds a converter for type_
            return partial(self.register_converter, type_=type_)

        if type_ is None:
            # get the type from the function annotation (the first argument to func)
            signature = inspect.signature(func)
            type_ = list(signature.parameters.values())[0].annotation

            # fail is there no annotation
            if type_ is inspect.Parameter.empty:
                raise ValueError(
                    "No type annotation for the first argument to the converter function. Please pass the "
                    "type as the type_ argument to register_converter or add a type annotation."
                )

            # check if type_ is Annoated
            if hasattr(type_, '__metadata__'):
                assert hasattr(type_, '__origin__')
                # get the actual type
                type_ = type_.__origin__  # type: ignore

            assert type_ is not None, "type_ is None"

        self.converters[type_] = func

        return func

    def add_converter(self, func: typing.Callable[[T, ObjectConverter], dict] | None = None):
        """
        Decorator that adds a converter to the ObjectConverter class that is wrapped.
        """

        def wrapper(type_: type):
            if func is None:
                # Check if type_ is a Pydantic model
                if isinstance(type_, type) and issubclass(type_, pydantic.BaseModel):
                    # Add a converter for the model
                    self.register_converter(convert_pydantic_model, type_)
                else:
                    raise ValueError("add_converter can only be used with Pydantic models.")

            # Add a converter for the type
            self.register_converter(func, type_)

            return type_

        return wrapper


def convert_pydantic_model(obj: pydantic.BaseModel, converter: ObjectConverter | None = None) -> dict:
    """
    Converts a pydantic model to a dict.
    """
    if converter is None:
        converter = convert_pydantic_model

    return {key: converter(value, converter) for key, value in obj.model_dump(mode='python').items()}



#  LLM Tracer
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import enum

from pydantic import BaseModel
from wandb.sdk.data_types import trace_tree

from llmtracer.frame_info import FrameInfo


class TraceNodeKind(str, enum.Enum):
    """
    The type of event.

    We match wandb's span kind for convienence and add more.
    """

    LLM = trace_tree.SpanKind.LLM
    CHAIN = trace_tree.SpanKind.CHAIN
    AGENT = trace_tree.SpanKind.AGENT
    TOOL = trace_tree.SpanKind.TOOL
    SCOPE = "SCOPE"
    CALL = "CALL"
    EVENT = "EVENT"

    def __repr__(self):
        return f"{self.__class__.__name__}.{self.name}"


class TraceNode(BaseModel):
    """
    Node in the trace tree.
    """

    kind: TraceNodeKind
    name: str | None
    event_id: int

    start_time_ms: int
    end_time_ms: int
    running: bool = False

    delta_frame_infos: list[FrameInfo]

    properties: dict[str, object]
    children: list['TraceNode']

    def collect_event_id_map(self, event_id_map=None) -> dict[int, 'TraceNode']:
        """
        Update a map from event id to node.
        """
        if event_id_map is None:
            event_id_map = {}

        event_id_map[self.event_id] = self
        for child in self.children:
            child.collect_event_id_map(event_id_map)
        return event_id_map

    def to_custom_dict(self, include_timing: bool = True, include_lineno: bool = True):
        custom_dict = {
            "kind": self.kind.value,
            "name": self.name,
            "event_id": self.event_id,
            "delta_frame_infos": [
                {
                    "module": frame_info.module,
                    "function": frame_info.function,
                    "code_context": frame_info.code_context,
                }
                | ({"lineno": frame_info.lineno} if include_lineno else {})
                for frame_info in self.delta_frame_infos
            ],
            "properties": self.properties,
            "children": [
                child.to_custom_dict(include_timing=include_timing, include_lineno=include_lineno)
                for child in self.children
            ],
        }
        if include_timing:
            custom_dict["start_time_ms"] = self.start_time_ms
            custom_dict["end_time_ms"] = self.end_time_ms

        return custom_dict


class Trace(BaseModel):
    """
    A trace tree.
    """

    name: str | None
    traces: list[TraceNode]
    properties: dict[str, object]
    unique_objects: dict[str, object]

    def build_event_id_map(self) -> dict[int, TraceNode]:
        """
        Build a map from event id to node.
        """
        event_id_map: dict[int, TraceNode] = {}
        for trace in self.traces:
            trace.collect_event_id_map(event_id_map)
        return event_id_map

    def to_custom_dict(self, include_timing: bool = True, include_lineno: bool = True):
        """
        Convert the trace to a jsonable format.

        Args:
            include_timing: Whether to include timing information.
            include_lineno: Whether to include line number information.

        Returns:
            The jsonable trace.
        """
        return {
            "name": self.name,
            "traces": [
                trace.to_custom_dict(include_timing=include_timing, include_lineno=include_lineno)
                for trace in self.traces
            ],
            "properties": self.properties,
            "unique_objects": self.unique_objects,
        }



#  LLM Tracer
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Inspired by https://stackoverflow.com/a/30024601/854731
from contextlib import AbstractContextManager
from timeit import default_timer


class StopwatchContext(AbstractContextManager):
    """A context the keeps track of the elapsed time.

    The elapsed time is available as the `elapsed_time` attribute.
    Nested contexts are not supported, but reusing the same context is.

    Example:
    ```
    with StopwatchContext() as stopwatch:
        ...

    print(stopwatch.elapsed_time)
    ```
    """

    def __init__(self):
        self._start_time = None
        self._elapsed_time = 0.0

    @property
    def elapsed_time(self):
        if self._start_time is not None:
            return self._elapsed_time + default_timer() - self._start_time
        return self._elapsed_time

    def __enter__(self):
        assert self._start_time is None
        self._start_time = default_timer()
        return super().__enter__()

    def __exit__(self, exc_type, exc_value, traceback):
        end_time = default_timer()
        self._elapsed_time += end_time - self._start_time
        self._start_time = None



#  LLM Tracer
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .convenience import add_event, build_trace, event_scope, register_object, update_event_properties, update_name
from .handlers.json_writer import JsonFileWriter
from .handlers.trace_viewer import TraceViewerIntegration
from .module_filtering import module_filter, module_filters
from .trace_builder import (
    TraceBuilder,
    TraceBuilderEventHandler,
    trace_calls,
    trace_module_filters,
    trace_object_converter,
)
from .trace_schema import Trace, TraceNode, TraceNodeKind
from .wandb_integration import wandb_build_trace_trees, wandb_tracer

__version__ = '1.2.1'



#  LLM Tracer
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import wandb

from llmtracer import TraceNodeKind, event_scope
from llmtracer.wandb_integration import wandb_tracer

wandb.init(project="blackboard-pagi", name="wandb_integration_test", reinit=True)  # type: ignore

with wandb_tracer("", module_filters="blackboard_pagi.*", stack_frame_context=0):
    with event_scope("foo", kind=TraceNodeKind.AGENT):
        with event_scope("bar", kind=TraceNodeKind.LLM):
            with event_scope("baz", kind=TraceNodeKind.CALL):
                pass
            with event_scope("baz2", kind=TraceNodeKind.CALL):
                pass
        with event_scope("bar2", kind=TraceNodeKind.LLM):
            with event_scope("baz", kind=TraceNodeKind.CALL):
                with event_scope("foo", kind=TraceNodeKind.AGENT):
                    with event_scope("bar", kind=TraceNodeKind.LLM):
                        with event_scope("baz", kind=TraceNodeKind.CALL):
                            pass
                        with event_scope("baz2", kind=TraceNodeKind.CALL):
                            pass
                    with event_scope("bar2", kind=TraceNodeKind.LLM):
                        with event_scope("baz", kind=TraceNodeKind.CALL):
                            with event_scope("foo", kind=TraceNodeKind.AGENT):
                                with event_scope("bar", kind=TraceNodeKind.LLM):
                                    with event_scope("baz", kind=TraceNodeKind.CALL):
                                        pass
                                    with event_scope("baz2", kind=TraceNodeKind.CALL):
                                        pass
                                with event_scope("bar2", kind=TraceNodeKind.LLM):
                                    with event_scope("baz", kind=TraceNodeKind.CALL):
                                        with event_scope("foo", kind=TraceNodeKind.AGENT):
                                            with event_scope("bar", kind=TraceNodeKind.LLM):
                                                with event_scope("baz", kind=TraceNodeKind.CALL):
                                                    pass
                                                with event_scope("baz2", kind=TraceNodeKind.CALL):
                                                    pass
                                            with event_scope("bar2", kind=TraceNodeKind.LLM):
                                                with event_scope("baz", kind=TraceNodeKind.CALL):
                                                    pass



"""See `TestDict` for an example."""
#  LLM Tracer
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import MutableMapping, Type

import pytest

from llmtracer.testing.collections import unordered_equal


class MutableMappingTests:
    mutable_mapping: Type

    @classmethod
    def create_mutable_mapping(cls) -> MutableMapping:
        return cls.mutable_mapping()

    @staticmethod
    def get_key(i):
        return i

    @staticmethod
    def get_value(i):
        return str(i)

    def get_key_value(self, i):
        return self.get_key(i), self.get_value(i)

    def test_get_missing_fails(self):
        instance = self.create_mutable_mapping()
        with pytest.raises(KeyError):
            # noinspection PyStatementEffect
            instance[self.get_key(1)]

    def test_del_missing_fails(self):
        instance = self.create_mutable_mapping()
        with pytest.raises(KeyError):
            del instance[self.get_key(1)]

    def test_integration(self):
        instance = self.create_mutable_mapping()

        assert len(instance) == 0
        assert list(iter(instance)) == []

        key, value = self.get_key_value(1)
        instance[key] = value

        assert instance[key] == value
        assert unordered_equal(list(iter(instance)), [key])
        assert len(instance) == 1

        key2, value2 = self.get_key_value(2)
        instance[key2] = value2

        assert instance[key2] == value2
        assert unordered_equal(list(iter(instance)), [key, key2])
        assert len(instance) == 2

        value3 = self.get_value(3)
        instance[key] = value3
        assert instance[key] == value3

        del instance[key]
        assert key not in instance
        assert unordered_equal(list(iter(instance)), [key2])
        assert len(instance) == 1



#  LLM Tracer
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import copy
import os
from dataclasses import dataclass
from enum import Enum

import svgwrite
from svgwrite.base import BaseElement
from svgwrite.elementfactory import factoryelements
from svgwrite.etree import etree
from svgwrite.mixins import Clipping, Presentation, Transform

from llmtracer.trace_builder import TraceBuilder, TraceBuilderEventHandler
from llmtracer.trace_schema import Trace, TraceNode, TraceNodeKind


# solarized colors as HTML hex
# https://ethanschoonover.com/solarized/
class SolarizedColors(str, Enum):
    base03 = "#002b36"
    base02 = "#073642"
    base01 = "#586e75"
    base00 = "#657b83"
    base0 = "#839496"
    base1 = "#93a1a1"
    base2 = "#eee8d5"
    base3 = "#fdf6e3"
    yellow = "#b58900"
    orange = "#cb4b16"
    red = "#dc322f"
    magenta = "#d33682"
    violet = "#6c71c4"
    blue = "#268bd2"
    cyan = "#2aa198"
    green = "#859900"


renderjson = """// Copyright © 2013-2017 David Caldwell <david@porkrind.org>
//
// Permission to use, copy, modify, and/or distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
// SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
// OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
// CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

// Usage
// -----
// The module exports one entry point, the `renderjson()` function. It takes in
// the JSON you want to render as a single argument and returns an HTML
// element.
//
// Options
// -------
// renderjson.set_icons("+", "-")
//   This Allows you to override the disclosure icons.
//
// renderjson.set_show_to_level(level)
//   Pass the number of levels to expand when rendering. The default is 0, which
//   starts with everything collapsed. As a special case, if level is the string
//   "all" then it will start with everything expanded.
//
// renderjson.set_max_string_length(length)
//   Strings will be truncated and made expandable if they are longer than
//   `length`. As a special case, if `length` is the string "none" then
//   there will be no truncation. The default is "none".
//
// renderjson.set_sort_objects(sort_bool)
//   Sort objects by key (default: false)
//
// renderjson.set_replacer(replacer_function)
//   Equivalent of JSON.stringify() `replacer` argument when it's a function
//
// renderjson.set_collapse_msg(collapse_function)
//   Accepts a function (len:number):string => {} where len is the length of the
//   object collapsed.  Function should return the message displayed when a
//   object is collapsed.  The default message is "X items"
//
// renderjson.set_property_list(property_list)
//   Equivalent of JSON.stringify() `replacer` argument when it's an array
//
// Theming
// -------
// The HTML output uses a number of classes so that you can theme it the way
// you'd like:
//     .disclosure    ("⊕", "⊖")
//     .syntax        (",", ":", "{", "}", "[", "]")
//     .string        (includes quotes)
//     .number
//     .boolean
//     .key           (object key)
//     .keyword       ("null", "undefined")
//     .object.syntax ("{", "}")
//     .array.syntax  ("[", "]")

var module, window, define, renderjson = (function() {
  var themetext = function( /* [class, text]+ */ ) {
    /* The arguments are alternating class names and text strings.
     *
     * This function creates a span using the class name and adds the text for each argument pair. It returns the array.
     */
    var spans = [];
    while (arguments.length)
      spans.push(append(span(Array.prototype.shift.call(
          arguments)),
        text(Array.prototype.shift.call(arguments))));
    return spans;
  };

  var append = function( /* el, ... */ ) {
    /* append all arguments to the first one and return the first el. */
    // Flattens arrays.
    var el = Array.prototype.shift.call(arguments);
    for (var a = 0; a < arguments.length; a++)
      if (arguments[a].constructor == Array)
        append.apply(this, [el].concat(arguments[a]));
      else
        el.appendChild(arguments[a]);
    return el;
  };

  var prepend = function(el, child) {
    /* prepend child to el's children */
    el.insertBefore(child, el.firstChild);
    return el;
  }

  var isempty = function(obj, pl) {
    var keys = pl || Object.keys(obj);
    for (var i in keys)
      if (Object.hasOwnProperty.call(obj, keys[i]))
        return false;
    return true;
  }

  var text = function(txt) {
    return document.createTextNode(txt)
  };

  var div = function() {
    return document.createElementNS(
      "http://www.w3.org/1999/xhtml", "div")
  };

  var pre = function(classname) {
    var s = document.createElementNS(
      "http://www.w3.org/1999/xhtml", "pre")
    if (classname) s.className = classname;
    return s;
  };

  var span = function(classname) {
    var s = document.createElementNS(
      "http://www.w3.org/1999/xhtml", "span");
    if (classname) s.className = classname;
    return s;
  };

  var A = function A(txt, classname, callback) {
    var a = document.createElementNS(
      "http://www.w3.org/1999/xhtml", "a");
    if (classname) a.className = classname;
    a.appendChild(text(txt));
    a.href = '#';
    a.onclick = function(e) {
      callback();
      if (e) e.stopPropagation();
      return false;
    };
    return a;
  };

  function _renderjson(json, indent, dont_indent, show_level,
    options) {
    var my_indent = dont_indent ? "" : indent;

    var disclosure = function(open, placeholder, close, type,
      builder) {
      var content;
      var empty = span(type);
      var show = function() {
        if (!content) {
          append(
            empty.parentNode,
            content = prepend(builder(),
              A(
                options.hide, "disclosure",
                function() {
                  content.style.display =
                    "none";
                  empty.style.display =
                    "inline";
                })
            )
          );
        }
        content.style.display = "inline";
        empty.style.display = "none";
      };
      append(empty,
        A(options.show, "disclosure", show),
        themetext(type + " syntax", open),
        A(placeholder, null, show),
        themetext(type + " syntax", close));

      var el = append(span(), text(my_indent.slice(0, -1)),
        empty);
      if (show_level > 0 && type != "string")
        show();
      return el;
    };

    if (json === null) return themetext(null, my_indent, "keyword",
      "null");
    if (json === void 0) return themetext(null, my_indent,
      "keyword", "undefined");

    if (typeof(json) == "string" && json.length > options
      .max_string_length)
      return disclosure('"', json.substr(0, options
          .max_string_length) + " ...", '"', "string",
        function() {
          return append(span("string"), themetext(null,
            my_indent, "string", JSON.stringify(
              json)));
        });

    // Strings: wrap in a <pre>
    if (json.constructor == String) {
        let pre_text = pre(null);
        pre_text.innerHTML = json;
        pre_text.style.whiteSpace = "pre-wrap";
        pre_text.style.display ="inline-flex";
        pre_text.style.overflow = "auto";
        pre_text.style.maxHeight = "400px";
        // add borders for overflow
        pre_text.style.border = "1px solid #ccc";
        // scrollbar-gutter: stable;
        pre_text.style.scrollbarGutter = "stable";
        pre_text.style.margin = "0";

        let as = [
            append(span(null), text(my_indent)),
            append(span("string"), text("\\""), pre_text, text("\\""))
        ];
        return as;
    }

    // Numbers and bools
    if (typeof(json) != "object" || [Number, Boolean, Date]
      .indexOf(json.constructor) >= 0)
      return themetext(null, my_indent, typeof(json), JSON
        .stringify(json));

    if (json.constructor == Array) {
      if (json.length == 0) return themetext(null, my_indent,
        "array syntax", "[]");

      return disclosure("[", options.collapse_msg(json.length),
        "]", "array",
        function() {
          var as = append(span("array"), themetext(
            "array syntax", "[", null, "\\n"));
          for (var i = 0; i < json.length; i++)
            append(as,
              _renderjson(options.replacer.call(json,
                  i, json[i]), indent + "    ",
                false, show_level - 1, options),
              i != json.length - 1 ? themetext(
                "syntax", ",") : [],
              text("\\n"));
          append(as, themetext(null, indent,
            "array syntax", "]"));
          return as;
        });
    }

    // object
    if (isempty(json, options.property_list))
      return themetext(null, my_indent, "object syntax", "{}");

    return disclosure("{", options.collapse_msg(Object.keys(json)
      .length), "}", "object", function() {
      var os = append(span("object"), themetext(
        "object syntax", "{", null, "\\n"));
      for (var k in json) var last = k;
      var keys = options.property_list || Object.keys(
        json);
      if (options.sort_objects)
        keys = keys.sort();
      for (var i in keys) {
        var k = keys[i];
        if (!(k in json)) continue;
        append(os, themetext(null, indent + "    ",
            "key", '"' + k + '"',
            "object syntax", ': '),
          _renderjson(options.replacer.call(json,
              k, json[k]), indent + "    ",
            true, show_level - 1, options),
          k != last ? themetext("syntax", ",") : [],
          text("\\n"));
      }
      append(os, themetext(null, indent, "object syntax",
        "}"));
      return os;
    });
  }

  var renderjson = function renderjson(json) {
    var options = new Object(renderjson.options);
    options.replacer = typeof(options.replacer) == "function" ?
      options.replacer : function(k, v) {
        return v;
      };
    var pre = append(document.createElementNS(
        "http://www.w3.org/1999/xhtml", "pre"),
      _renderjson(json, "", false, options.show_to_level,
        options));
    pre.className = "renderjson";
    return pre;
  }
  renderjson.set_icons = function(show, hide) {
    renderjson.options.show = show;
    renderjson.options.hide = hide;
    return renderjson;
  };
  renderjson.set_show_to_level = function(level) {
    renderjson.options.show_to_level = typeof level ==
      "string" &&
      level.toLowerCase() === "all" ? Number.MAX_VALUE :
      level;
    return renderjson;
  };
  renderjson.set_max_string_length = function(length) {
    renderjson.options.max_string_length = typeof length ==
      "string" &&
      length.toLowerCase() === "none" ? Number.MAX_VALUE :
      length;
    return renderjson;
  };
  renderjson.set_sort_objects = function(sort_bool) {
    renderjson.options.sort_objects = sort_bool;
    return renderjson;
  };
  renderjson.set_replacer = function(replacer) {
    renderjson.options.replacer = replacer;
    return renderjson;
  };
  renderjson.set_collapse_msg = function(collapse_msg) {
    renderjson.options.collapse_msg = collapse_msg;
    return renderjson;
  };
  renderjson.set_property_list = function(prop_list) {
    renderjson.options.property_list = prop_list;
    return renderjson;
  };
  // Backwards compatiblity. Use set_show_to_level() for new code.
  renderjson.set_show_by_default = function(show) {
    renderjson.options.show_to_level = show ? Number.MAX_VALUE :
      0;
    return renderjson;
  };
  renderjson.options = {};
  renderjson.set_icons('⊕', '⊖');
  renderjson.set_show_by_default(false);
  renderjson.set_sort_objects(false);
  renderjson.set_max_string_length("none");
  renderjson.set_replacer(void 0);
  renderjson.set_property_list(void 0);
  renderjson.set_collapse_msg(function(len) {
    return len + " item" + (len == 1 ? "" : "s")
  })
  return renderjson;
})();

if (define) define({
  renderjson: renderjson
})
else(module || {}).exports = (window || {}).renderjson = renderjson;"""


def convert_trace_node_kind_to_color(kind: TraceNodeKind):
    if kind == TraceNodeKind.SCOPE:
        return SolarizedColors.base1
    elif kind == TraceNodeKind.AGENT:
        return SolarizedColors.green
    elif kind == TraceNodeKind.LLM:
        return SolarizedColors.blue
    elif kind == TraceNodeKind.CHAIN:
        return SolarizedColors.cyan
    elif kind == TraceNodeKind.CALL:
        return SolarizedColors.yellow
    elif kind == TraceNodeKind.EVENT:
        return SolarizedColors.orange
    elif kind == TraceNodeKind.TOOL:
        return SolarizedColors.magenta
    else:
        return SolarizedColors.base2


class ForeignObject(BaseElement, Transform, Presentation, Clipping):
    """ForeignObject element.

    The **rect** element defines a rectangle which is axis-aligned with the current
    user coordinate system. Rounded rectangles can be achieved by setting appropriate
    values for attributes **rx** and **ry**.
    """

    elementname = 'foreignObject'
    inner_xml: etree.Element

    def __init__(self, inner_xml: etree.Element, insert=None, size=None, **extra):
        """Initialize a new instance.

        :param inner_xml: XML `ElementTree` of the inner content
        :param 2-tuple insert: insert point (**x**, **y**), left-upper point
        :param 2-tuple size: (**width**, **height**)
        :param extra: additional SVG attributes as keyword-arguments

        """
        super(ForeignObject, self).__init__(**extra)
        if insert is not None:
            self['x'] = insert[0]
            self['y'] = insert[1]
        if size is not None:
            self['width'] = size[0]
            self['height'] = size[1]
        self.inner_xml = inner_xml

    def get_xml(self):
        """Get the XML representation as `ElementTree` object.

        :return: XML `ElementTree` of this object and all its subelements

        """
        xml = etree.Element(self.elementname)
        if self.debug:
            self.validator.check_all_svg_attribute_values(self.elementname, self.attribs)
        for attribute, value in sorted(self.attribs.items()):
            # filter 'None' values
            if value is not None:
                value = self.value_to_string(value)
                if value:  # just add not empty attributes
                    xml.set(attribute, value)

        for element in self.elements:
            xml.append(element.get_xml())
        xml.append(self.inner_xml)
        return xml


factoryelements['foreignObject'] = ForeignObject


def create_svg_from_trace(trace: Trace):
    total_width = 1280
    total_height = 760
    dwg = svgwrite.Drawing(
        'trace.svg',
        profile='full',
        size=(total_width, total_height),
        style="background-color: " + SolarizedColors.base3.value,
        debug=False,
    )

    dwg.add(
        dwg.script(
            type="application/ecmascript",
            content=renderjson,
        )
    )
    # Add the JavaScript element
    dwg.add(
        dwg.script(
            type="application/ecmascript",
            content="""
                function handleClick(evt) {
                    // implement the zooming functionality and display the rectangle's properties.
                    let target = evt.target;
                    // get the parent of the target
                    let parent = target.parentNode;
                    // get x of parent
                    let left = parent.x.baseVal.value;
                    // get y of parent
                    let top = parent.y.baseVal.value;
                    // set the view_zoom's href to this id
                    let zoom_view = document.getElementById("zoom_view")
                    // if zoom_view is already set to the parent.id, reset to the full_view
                    if (zoom_view.getAttribute("href") == "#" + parent.id) {
                        zoom_view.setAttribute("href", "#full_view");
                        zoom_view.setAttribute("x", 0);
                        zoom_view.setAttribute("y", 0);
                        resetDetails();
                    } else {
                        zoom_view.setAttribute("href", "#" + parent.id);
                        // set the view_zoom's x and y to the negative x and y of the bounding box
                        zoom_view.setAttribute("x", -left);
                        zoom_view.setAttribute("y", -top);
                        updateDetails(parent);
                    }
                }
                """,
        )
    )

    # New code:
    # Create an icicle plot from the trace data.
    # Use the start time of the first node as the start time of the trace.
    start_time = trace.traces[0].start_time_ms
    end_time = trace.traces[-1].end_time_ms

    def traverse_node(node: TraceNode, parent, level: int, parent_start_time_ms: float, parent_duration_ms: float):
        # create a group for node
        node_group = dwg.svg(
            id=str(node.event_id),
            x=(node.start_time_ms - parent_start_time_ms) / parent_duration_ms * 99.5 * svgwrite.percent,
            y="0" if level == 0 else "1.8em",
            width=(node.end_time_ms - node.start_time_ms)
            * 0.99**level
            / parent_duration_ms
            * 99.5
            * svgwrite.percent,
        )
        node_copy = copy.deepcopy(node)
        node_copy.children = []
        node_group["data-raw"] = node_copy.json()
        parent.add(node_group)

        rect = dwg.rect(
            insert=("2px", "0"),
            size=("99.5%", "1.7em"),
            fill=convert_trace_node_kind_to_color(node.kind).value,
            onclick="handleClick(evt)",
            id=f"{str(node.event_id)}_box",
            rx=2,
            ry=2,
        )
        rect["data-id"] = str(node.event_id)
        node_group.add(rect)

        clip_path = dwg.clipPath(id=f"{str(node.event_id)}_clip")
        clip_path.add(
            dwg.rect(
                insert=("4px", "0"),
                size=("99.5%", "1.7em"),
                fill="white",
            )
        )
        dwg.defs.add(clip_path)

        # add a text element for the name of the node
        text = dwg.text(
            node.name,
            insert=(
                "4px",
                "1.35em",
            ),
            fill="black",
            font_size="1em",
            font_family="Arial",
            clip_path=clip_path.get_funciri(),
            pointer_events="none",
        )
        node_group.add(text)

        # traverse children
        for child in node.children:
            traverse_node(child, node_group, level + 1, node.start_time_ms, node.end_time_ms - node.start_time_ms)

    symbol = dwg.symbol(id="full_view")
    dwg.defs.add(symbol)

    for node in trace.traces:
        traverse_node(node, symbol, level=0, parent_start_time_ms=start_time, parent_duration_ms=end_time - start_time)

    zoom_use = dwg.use(id='zoom_view', x=0, y=0, width="100%", height=total_height - 560, href=symbol.get_iri())
    dwg.add(zoom_use)

    details_pane = dwg.foreignObject(
        x=0,
        y=total_height - 560,
        width="100%",
        height=560,
        inner_xml=etree.fromstring(
            """
<html xmlns="http://www.w3.org/1999/xhtml" style="height: 100%;">
<body style="height: 100%; margin: 0px; padding: 2px;">
<div style="height: 100%; display: flex; flex-direction: column; background-color: #eee8d5;">
    <style>
        // Solarized Light Color Scheme (Different Colors for Different Types)
        .renderjson a              { text-decoration: none; }
        .renderjson .syntax        { color: #657b83; }
        .renderjson .string        { color: #2aa198; }
        .renderjson .number        { color: #d33682; }
        .renderjson .boolean       { color: #6c71c4; }
        .renderjson .key           { color: #268bd2; }
        .renderjson .keyword       { color: #859900; }
        .renderjson .object.syntax { color: #b58900; }
        .renderjson .array.syntax  { color: #cb4b16; }
        .renderjson .disclosure    { color: #b58900;
                                     font-size: 150%; }
        .renderjson .syntax.collapsed .disclosure { color: #cb4b16; }
        .renderjson .disclosure:hover    { color: #dc322f; }
        .renderjson .syntax.collapsed .disclosure:hover { color: #d33682; }
        // background for the div
        div {
            color: #657b83;
            font-family: monospace;
            font-size: 14px;
            overflow: auto;
        }
    </style>
    <strong>Click on a node to zoom in and see more details. Click on the zoomed in node to zoom out.</strong>
    <h2>Scope Details</h2>
    <div id="details" style="overflow: auto; font-family: monospace;">
        <table style="border: 1px none black; border-collapse: collapse; width: 100%;">
            <tr style="border: 1px none black; border-collapse: collapse;">
                <th style="border: 1px none black; border-collapse: collapse; text-align: left; width: 30ex;">
                    Property
                </th>
                <th style="border: 1px none black; border-collapse: collapse; text-align: left;"></th>
            </tr>
            <tr style="border: 1px none black; border-collapse: collapse;">
                <td style="border: 1px none black; border-collapse: collapse; text-align: left; width: 30ex;">Name</td>
                <td id="details-name" style="border: 1px none black; border-collapse: collapse; text-align: left;"></td>
            </tr>
            <tr style="border: 1px none black; border-collapse: collapse;">
                <td style="border: 1px none black; border-collapse: collapse; text-align: left; width: 30ex;">
                Kind</td>
                <td id="details-kind" style="border: 1px none black; border-collapse: collapse; text-align: left;"></td>
            </tr>
            <tr style="border: 1px none black; border-collapse: collapse;">
                <td style="border: 1px none black; border-collapse: collapse; text-align: left; width: 30ex;">
                Duration</td>
                <td id="details-duration"
                style="border: 1px none black; border-collapse: collapse; text-align: left;"></td>
            </tr>
            <tr style="border: 1px none black; border-collapse: collapse;">
                <td style="border: 1px none black; border-collapse: collapse; text-align: left; vertical-align: top;
                    width: 30ex;">Exception</td>
                <td id="details-exception"
                style="border: 1px none black; border-collapse: collapse; text-align: left; width: calc(100%-30ex);">
                </td>
            </tr>
            <tr style="border: 1px none black; border-collapse: collapse;">
                <td style="border: 1px none black; border-collapse: collapse; text-align: left; vertical-align: top;
                    width: 30ex;">Result</td>
                <td id="details-result"
                style="border: 1px none black; border-collapse: collapse; text-align: left; width: calc(100%-30ex);">
                </td>
            </tr>
            <tr style="border: 1px none black; border-collapse: collapse;">
                <td style="border: 1px none black; border-collapse: collapse; text-align: left; vertical-align: top;
                    width: 30ex;">Arguments</td>
                <td id="details-arguments"
                style="border: 1px none black; border-collapse: collapse; text-align: left; width: calc(100%-30ex);">
                </td>
            </tr>
            <tr style="border: 1px none black; border-collapse: collapse;">
                <td style="border: 1px none black; border-collapse: collapse; text-align: left; vertical-align: top;
                    width: 30ex;">Self</td>
                <td id="details-self"
                style="border: 1px none black; border-collapse: collapse; text-align: left; width: calc(100%-30ex);">
                </td>
            </tr>
            <tr style="border: 1px none black; border-collapse: collapse;">
                <td style="border: 1px none black; border-collapse: collapse; text-align: left; vertical-align: top;
                    width: 30ex;">Properties</td>
                <td id="details-properties"
                style="border: 1px none black; border-collapse: collapse; text-align: left; width: calc(100%-30ex);">
                </td>
            </tr>
        </table>
    </div>
</div>
</body></html>
            """
        ),
    )
    dwg.add(details_pane)

    dwg.add(
        dwg.script(
            type="application/ecmascript",
            content="""
        function updateDetails(target) {
            let trace_info = JSON.parse(target.dataset.raw);
            let details_name = document.getElementById("details-name");
            let details_kind = document.getElementById("details-kind");
            let details_duration = document.getElementById("details-duration");
            let details_exception = document.getElementById("details-exception");
            let details_result = document.getElementById("details-result");
            let details_arguments = document.getElementById("details-arguments");
            let details_self = document.getElementById("details-self");
            let details_properties = document.getElementById("details-properties");

            var node_name = trace_info.name || "/Unnamed/";
            // add an asterisk if the node is currently running (.running is True)
            if (trace_info.running) {
                node_name += " (*)";
            }
            details_name.innerText = node_name;

            // set kind
            details_kind.innerText = trace_info.kind;

            // set duration
            details_duration.innerText = ((trace_info.end_time_ms - trace_info.start_time_ms) / 1000).toFixed(2) + " s";

            function renderjson_to(node, json) {
                node.innerHTML = "";
                rendered_json = renderjson.set_icons('+', '-').set_show_to_level("all")(json)
                rendered_json.style.whiteSpace = "pre-wrap";
                rendered_json.style.wordWrap = "break-word";
                rendered_json.style.overflow = "auto";
                rendered_json.style.display = "inline-block";
                rendered_json.style.margin = "0";
                node.appendChild(rendered_json);
            }

            // set exception
            let exception = trace_info.properties.exception || "";
            renderjson_to(details_exception, exception);

            // set result
            let result = trace_info.properties.result || "";
            renderjson_to(details_result, result);

            // set arguments
            let arguments = trace_info.properties.arguments || {};
            renderjson_to(details_arguments, arguments);

            // set self
            let self = trace_info.properties.self || {};
            renderjson_to(details_self, self);

            // set properties
            // remove exception, result, arguments, self
            delete trace_info.properties.exception;
            delete trace_info.properties.result;
            delete trace_info.properties.arguments;
            delete trace_info.properties.self;
            let properties = trace_info.properties || {};
            renderjson_to(details_properties, properties);
        }

        function resetDetails() {
            let details_name = document.getElementById("details-name");
            let details_kind = document.getElementById("details-kind");
            let details_duration = document.getElementById("details-duration");
            let details_exception = document.getElementById("details-exception");
            let details_result = document.getElementById("details-result");
            let details_arguments = document.getElementById("details-arguments");
            let details_self = document.getElementById("details-self");
            let details_properties = document.getElementById("details-properties");

            details_name.innerText = "";
            details_kind.innerText = "";
            details_duration.innerText = "";
            details_exception.innerHTML = "";
            details_result.innerHTML = "";
            details_arguments.innerHTML = "";
            details_self.innerHTML = "";
            details_properties.innerHTML = "";
        }
        """,
        )
    )

    return dwg


def save_trace_as_svg(filename: str, trace: Trace):
    tempfile = filename + ".new_tmp"
    svg = create_svg_from_trace(trace)
    svg.saveas(tempfile)
    os.replace(tempfile, filename)


@dataclass
class SvgFileWriter(TraceBuilderEventHandler):
    filename: str

    def on_scope_final(self, builder: 'TraceBuilder'):
        save_trace_as_svg(self.filename, builder.build())


# main
if __name__ == "__main__":
    # load example trace data
    trace = Trace.parse_file("spikes/optimization_unit_trace_ada_2023-05-19_11-07-07.json")
    save_trace_as_svg("trace.svg", trace)



#  LLM Tracer
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import os
import sys

import reflex.reflex

# Update the PYTHONPATH and change directories into the app directory
# This is necessary for the import below to work.
# Pynecone expects to be run from the root directory of the app.
sys.path.append(os.path.abspath(os.path.dirname(__file__)))
os.chdir(os.path.dirname(__file__))


def main():
    """This is the entry point for the CLI."""
    print(os.getcwd())
    reflex.reflex._run()


if __name__ == "__main__":
    main()



#  LLM Tracer
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass

from llmtracer.utils.callable_wrapper import CallableWrapper


def test_wrapper_functor():
    @dataclass
    class MyFunctor(CallableWrapper):
        __wrapped__: object

        def __call__(self, *args, **kwargs):
            return self.__wrapped__(*args, **kwargs)

    def my_functor(x: int) -> str:
        return str(x)

    my_functor = MyFunctor(my_functor)
    assert my_functor(3) == "3"

    # test that we can access the wrapped function's attributes
    @dataclass
    class InnerFunctor:
        a: int
        b: str

        def __call__(self, x: int) -> str:
            return str(x + self.a) + self.b

    my_functor = MyFunctor(InnerFunctor(3, "foo"))

    assert my_functor(3) == "6foo"
    assert my_functor.a == 3
    assert my_functor.b == "foo"

    # assert that the dir of the wrapper contains all attributes of the wrapped object (superset)
    assert set(dir(my_functor)) >= set(dir(my_functor.__wrapped__))


try:
    from pydantic import BaseModel

    def test_callable_wrapper_within_pydantic():
        # Pydantic seems to be doing weird things
        @dataclass
        class MyFunctor(CallableWrapper):
            __wrapped__: object

            def __call__(self, *args, **kwargs):
                return self.__wrapped__(*args, **kwargs)

        class PydanticModel(BaseModel):
            @MyFunctor
            def test(self, a):
                return a

        model = PydanticModel()
        assert model

        assert model.test("foo") == "foo"

except ImportError:
    pass



#  LLM Tracer
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from llmtracer import build_trace, event_scope, trace_calls


def test_trace():
    with build_trace(module_filters=__name__, stack_frame_context=0).scope() as builder:
        with event_scope("foo"):
            with event_scope("bar"):
                with event_scope("baz"):
                    pass

    assert builder is not None
    assert builder.build().to_custom_dict(include_timing=False, include_lineno=False)['traces'] == [
        {
            'children': [
                {
                    'children': [
                        {
                            'children': [
                                {
                                    'children': [],
                                    'delta_frame_infos': [],
                                    'event_id': 4,
                                    'kind': 'SCOPE',
                                    'name': 'baz',
                                    'properties': {},
                                }
                            ],
                            'delta_frame_infos': [],
                            'event_id': 3,
                            'kind': 'SCOPE',
                            'name': 'bar',
                            'properties': {},
                        }
                    ],
                    'delta_frame_infos': [],
                    'event_id': 2,
                    'kind': 'SCOPE',
                    'name': 'foo',
                    'properties': {},
                }
            ],
            'delta_frame_infos': [],
            'event_id': 1,
            'kind': 'SCOPE',
            'name': None,
            'properties': {},
        }
    ]


def test_trace_calls():
    @trace_calls(capture_args=True, capture_return=True)
    def f(value: int):
        return value * 3

    with build_trace(
        module_filters=__name__,
        stack_frame_context=0,
    ).scope() as builder:
        f(3)
        f(5)

    assert builder is not None
    traces = builder.build().to_custom_dict(include_timing=False, include_lineno=False)['traces']
    assert traces == [
        {
            'children': [
                {
                    'children': [],
                    'delta_frame_infos': [],
                    'event_id': 2,
                    'kind': 'CALL',
                    'name': 'f',
                    'properties': {'arguments': {'value': 3}, 'result': 9},
                },
                {
                    'children': [],
                    'delta_frame_infos': [],
                    'event_id': 3,
                    'kind': 'CALL',
                    'name': 'f',
                    'properties': {'arguments': {'value': 5}, 'result': 15},
                },
            ],
            'delta_frame_infos': [],
            'event_id': 1,
            'kind': 'SCOPE',
            'name': None,
            'properties': {},
        }
    ]



#  LLM Tracer
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import time
import typing
from dataclasses import dataclass

import requests

from llmtracer.utils.weakrefs import WeakKeyIdMap

from . import pcconfig

if typing.TYPE_CHECKING:
    from llmtracer import TraceBuilder


@dataclass
class TraceBuilderUpdate:
    token: str
    last_sent_ms: int


# Weak key dictionary from trace to TraceUpdates
_trace_builder_updates: WeakKeyIdMap['TraceBuilder', TraceBuilderUpdate] = WeakKeyIdMap()

_MIN_SEND_INTERVAL_MS = 500


def trace_viewer_send_trace_builder(trace_builder: 'TraceBuilder', force: bool = False):
    """Send a trace to the backend.

    :param trace_buider: The trace to send.
    :param force: If True, send the trace even if it has not been updated.
    """
    # Get the TraceUpdates object for this trace
    trace_updates = _trace_builder_updates.get(trace_builder)
    trace = trace_builder.build()
    if trace_updates is None:
        if trace.name is not None:
            # Use the trace name as the token if it is set
            token = trace.name
        else:
            # Generate a random token based on the current time
            token = str(int(time.time()))

        # Create a new TraceUpdates object if one does not exist
        trace_updates = TraceBuilderUpdate(token=token, last_sent_ms=-_MIN_SEND_INTERVAL_MS)
        _trace_builder_updates[trace_builder] = trace_updates

    now_ms = int(time.time() * 1000)
    if not force:
        # Check if the trace has been updated since the last send
        if now_ms - trace_updates.last_sent_ms < _MIN_SEND_INTERVAL_MS:
            return
    trace_updates.last_sent_ms = now_ms

    url = pcconfig.config.api_url + "/trace/" + trace_updates.token
    payload = trace.dict()
    try:
        requests.post(url, json=payload, timeout=0.05)
    except requests.exceptions.RequestException:
        pass



#  LLM Tracer
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import gc
from dataclasses import dataclass

import pytest

from llmtracer.testing.collections.mutable_mapping import MutableMappingTests
from llmtracer.testing.collections.mutable_set import MutableSetTests
from llmtracer.utils import weakrefs


class DummySupportsWeakRefs:
    pass


class DummyDoesNotSupportWeakRefs:
    __slots__ = ()


@dataclass
class DummyDataclass:
    field: str


def test_supports_weakref():
    assert not weakrefs.supports_weakrefs(1)
    # This is slightly surprising to be honest...
    assert not weakrefs.supports_weakrefs(object())

    assert weakrefs.supports_weakrefs(DummySupportsWeakRefs())
    assert not weakrefs.supports_weakrefs(DummyDoesNotSupportWeakRefs())

    assert weakrefs.supports_weakrefs(DummyDataclass("Hello"))


def test_id_map_finalizer():
    id_map_finalizer = weakrefs.IdMapFinalizer()

    a = DummySupportsWeakRefs()

    a_has_been_finalized_counter = 0
    id_a = id(a)

    def custom_handler(id_value):
        nonlocal a_has_been_finalized_counter
        assert id_value == id_a
        a_has_been_finalized_counter += 1

    id_map_finalizer.register(a, custom_handler)

    with pytest.raises(ValueError):
        id_map_finalizer.register(a, None)

    # Don't throw and don't do anything when we add a value that can't be weakref'ed.
    id_map_finalizer.register(1, None)

    assert set(id_map_finalizer) == {a}

    id_map_finalizer.release(a)

    assert a_has_been_finalized_counter == 0

    id_map_finalizer.register(a, custom_handler)
    gc.collect()

    assert id_map_finalizer.get_object(id(a)) == a

    assert a_has_been_finalized_counter == 0

    a = None
    gc.collect()

    assert a_has_been_finalized_counter == 1


def test_weak_key_id_map():
    weak_key_id_map = weakrefs.WeakKeyIdMap()

    a = DummySupportsWeakRefs()
    b = DummySupportsWeakRefs()

    weak_key_id_map[a] = 1
    weak_key_id_map[b] = 2

    gc.collect()

    assert a in weak_key_id_map
    assert b in weak_key_id_map

    assert weak_key_id_map[a] == 1
    assert weak_key_id_map[b] == 2

    a = None
    gc.collect()

    assert a not in weak_key_id_map
    assert weak_key_id_map == {b: 2}

    b = None
    gc.collect()

    assert not weak_key_id_map


@dataclass
class BoxedValue:
    i: int


class TestWeakKeyIdMap(MutableMappingTests):
    mutable_mapping = weakrefs.WeakKeyIdMap

    @staticmethod
    def get_key(i):
        return BoxedValue(i)


class TestWrappedValueMutableMapping(MutableMappingTests):
    class WrappedValueMutableMappingInstance(weakrefs.AbstractWrappedValueMutableMapping):
        @staticmethod
        def value_to_store(v):
            return (v,)

        @staticmethod
        def store_to_value(v):
            return v[0]

    @staticmethod
    def create_mutable_mapping():
        return TestWrappedValueMutableMapping.WrappedValueMutableMappingInstance()


class TestWeakIdSet(MutableSetTests):
    mutable_set = weakrefs.WeakIdSet

    @staticmethod
    def get_element(i):
        return BoxedValue(i)
